@@ -1,8 +1,8 @@
 #!/usr/bin/env ruby
 #
 # Samizdat member login
-# 
-#   Copyright (c) 2002-2003 Dmitry Borodaenko <angdraug@debian.org>
+#
+#   Copyright (c) 2002-2004  Dmitry Borodaenko <angdraug@debian.org>
 #
 #   This program is free software.
 #   You can distribute/modify this program under the terms of
@@ -18,5 +18,5 @@ login, passwd = session.params %w[login passwd]
 if login and passwd then
     session.open login, passwd
 else
-    session.out({'status' => 'REDIRECT', 'location' => session.base})
+    session.redirect(session.base)
 end
