@@ -1,8 +1,8 @@
 #!/usr/bin/env ruby
 #
 # Samizdat logout
-# 
-#   Copyright (c) 2002-2003 Dmitry Borodaenko <angdraug@debian.org>
+#
+#   Copyright (c) 2002-2004  Dmitry Borodaenko <angdraug@debian.org>
 #
 #   This program is free software.
 #   You can distribute/modify this program under the terms of
@@ -14,4 +14,4 @@ require 'samizdat'
 session = Session.new
 
 session.close if session.id
-session.out({'status' => 'REDIRECT', 'location' => session.referer})
+session.redirect
