@@ -1,6 +1,6 @@
 # Samizdat RDF storage
-# 
-#   Copyright (c) 2002-2003 Dmitry Borodaenko <angdraug@debian.org>
+#
+#   Copyright (c) 2002-2004  Dmitry Borodaenko <angdraug@debian.org>
 #
 #   This program is free software.
 #   You can distribute/modify this program under the terms of
@@ -76,7 +76,7 @@ class SamizdatRDF
     # get all query answers (similar to DBI#select_all)
     #
     def select_all(query, limit=nil, offset=nil, &p)
-        sql = select(query)
+        sql = select(query).dup
         sql << "\nLIMIT #{limit}" if limit
         sql << "\nOFFSET #{offset}" if offset
         if block_given? then
@@ -86,54 +86,32 @@ class SamizdatRDF
         end
     end
 
-    # translate Squish SELECT query to SQL
-    #
-    # accepts String or pre-parsed SquishQuery object
+    @@select_cache = {}
+
+    # accepts String or pre-parsed SquishQuery object, caches SQL by String
     #
     def select(query)
         case query
         when String
-            q = SquishQuery.new(self, query)
+            return @@select_cache[query] if @@select_cache.has_key? query
+            @@select_cache = {} if @@select_cache.size >= 1024
+            @@select_cache[query] = SquishQuery.new(self, query).to_sql
         when SquishQuery
-            q = query
+            query.to_sql
         else
             raise ProgrammingError, "String or SquishQuery expected"
         end
-
-        raise ProgrammingError, "Wrong query type: select expected" unless
-            q.type == :select
-
-        literal, order, where = q.literal, q.order, q.jc
-        q.pm.each_key do |node|
-            re = Regexp.new('(^|[\s\(\)])' + Regexp.escape(node) + \
-                '([\s\(\)]|$)')
-            literal = literal.gsub(re, ' '+q.b[node]+' ')
-            order = order.gsub(re, ' '+q.b[node]+' ')
-        end
-        where.push '(' + literal + ')' unless literal == ''
-        order = "\nORDER BY " + order + ' ' + q.order_dir unless order == ''
-
-        # todo: make SELECT DISTINCT work
-        result = q.unstring('SELECT ' +
-            q.nodes.collect {|node|
-                q.b[node] + ' AS ' + node.sub(SquishQuery::BN) { $1 }
-            }.compact.join(', ') +
-            "\nFROM " + q.table.collect {|a,t| t + ' ' + a }.join(', ') +
-            ((where.length > 0)? "\nWHERE " + where.join("\n  AND ") : '') +
-            order)
-        raise ProgrammingError, "Unexpected '?' in translated query (probably, caused by unmapped blank node): #{result.gsub(/\s+/, ' ')};" if result =~ /\?/
-        return result
     end
 
     # merge Squish query into RDF database
     #
-    def merge(query)
+    def assert(query)
         q = SquishQuery.new(self, query)
-        raise ProgrammingError, 'Wrong query type: merge expected' unless
-            q.type == :merge
+        raise ProgrammingError, 'Wrong query type: assert expected' unless
+            q.type == :assert
         insert, update = q.nodes
 
-        @db.transaction do |db|   # AutoCommit assumed off
+@db.transaction do |db|   # AutoCommit assumed off
 
         # Stage 1: Resources
         v = {}   # v: node -> value
@@ -200,7 +178,7 @@ class SamizdatRDF
             end
         end
 
-        end   # transaction
+end   # transaction
 
         return nil
     end
@@ -236,8 +214,26 @@ class SquishQuery
     # regexp for blank node mark and name
     BN = Regexp.new(/\A\?([[:alnum:]_]+)\z/).freeze
 
-    # regexp for string literal
-    LITERAL = Regexp.new(/\A\$\{\d+\}\z/).freeze
+    # regexp for replaced string literal
+    LITERAL = Regexp.new(/\A'\d+'\z/).freeze
+
+    # regexp for number
+    NUMBER = Regexp.new(/\A-?[[:digit:]]+(\.[[:digit:]]+)?\z/).freeze
+
+    # regexp for operator
+    OPERATOR = Regexp.new(/\A(\+|-|\*|\/|<|<=|>|>=|=|I?LIKE|AND|OR)\z/i).freeze
+
+    # regexp for aggregate function
+    AGGREGATE = Regexp.new(/\A(avg|count|max|min|sum)\z/i).freeze
+
+    QUERY = Regexp.new(/\A\s*(SELECT|INSERT|UPDATE)\b\s*(.*?)\s*
+        \bWHERE\b\s*(.*?)\s*
+        (?:\bLITERAL\b\s*(.*?))?\s*
+        (?:\bGROUP\s+BY\b\s*(.*?))?\s*
+        (?:\bORDER\s+BY\b\s*(.*?)\s*(ASC|DESC)?)?\s*
+        (?:\bLIMIT\b\s*(.*?))?\s*
+        (?:\bOFFSET\b\s*(.*?))?\s*
+        (?:\bUSING\b\s*(.*?))?\s*\z/mix).freeze
 
     # extract common Squish query sections, perform namespace substitution,
     # generate query pattern graph, call transform_pattern,
@@ -255,20 +251,18 @@ class SquishQuery
                 "Bad query initialization parameter class: #{query.class}"
         end
 
-        # replace string literals with ${n} placeholders
+        # replace string literals with 'n' placeholders
         @strings = []
-        query = query.gsub(/'(''|[^'])*'/m) do
+        query.gsub!(/'(''|[^'])*'/m) do
             @strings.push $&
-            '${' + (@strings.length - 1).to_s + '}'
+            "'" + (@strings.size - 1).to_s + "'"
         end
 
-        query =~ /\A\s*(SELECT|INSERT|UPDATE)\b\s*(.*?)\s*
-            \bWHERE\b\s*(.*?)\s*
-            (?:\bLITERAL\b\s*(.*?))?\s*
-            (?:\bORDER\s+BY\b\s*(.*?)\s*(ASC|DESC)?)?\s*
-            (?:\bUSING\b\s*(.*?))?\s*\z/mix or raise ProgrammingError, "Malformed query: are keywords SELECT, INSERT, UPDATE or WHERE missing?"
-        @key, @nodes, @pattern, @literal, @order, @order_dir, @ns =
-            $1.upcase, $2, $3, $4.to_s, $5.to_s, $6.to_s.upcase, $7.to_s
+        match = QUERY.match(query) or raise ProgrammingError, "Malformed query: are keywords SELECT, INSERT, UPDATE or WHERE missing?"
+        match, @key, @nodes, @pattern, @literal, @group, @order, @order_dir,
+            @limit, @offset, @ns = match.to_a.collect {|m| m.to_s }
+        @key.upcase!
+        @order_dir.upcase!
 
         # todo: validate ns
         @ns = (@ns =~ /\APRESET\s+NS\z/) ? config['ns'] :
@@ -282,20 +276,25 @@ class SquishQuery
             end
         end
         validate_expression(@literal)
+        @group.split(/\s*,\s*/).each {|group| validate_expression(group) }
         validate_expression(@order)
+        raise "LIMIT and OFFSET must be numbers" unless
+            (@limit == '' or @limit =~ NUMBER) and
+            (@offset == '' or @offset =~ NUMBER)
 
         transform_pattern
 
+        query.scan(/\?\b.+?\b/) do |node|
+            raise ProgrammingError, "Blank node '#{node}' is not bound by the query pattern" unless @b[node]
+        end
+
         # determine query type, parse and validate nodes section
         if @key == 'SELECT' then
             @type = :select
-            @nodes = @nodes.split(/,\s*/)
-            @nodes.each do |node|
-                raise ProgrammingError, "Blank node expected in must-bind (SELECT) section instead of '#{node}'" unless node =~ BN
-                raise ProgrammingError, "Blank node '#{node}' is not bound by the query pattern" unless @b[node]
-            end
+            @nodes = @nodes.split(/\s*,\s*/)
+            @nodes.each {|node| validate_expression(node) }
         else
-            @type = :merge
+            @type = :assert
             if @key == 'UPDATE' then
                 insert = ''
                 update = @nodes
@@ -306,10 +305,10 @@ class SquishQuery
                 raise ProgrammingError, \
                     "Query doesn't start with one of SELECT, INSERT, or UPDATE"
             end
-            insert = insert.split(/,\s*/).each {|s|
+            insert = insert.split(/\s*,\s*/).each {|s|
                 raise ProgrammingError, "Blank node expected in INSERT section instead of '#{s}'" unless s =~ BN
             }
-            update = Hash[*update.split(/,\s*/).collect {|s|
+            update = Hash[*update.split(/\s*,\s*/).collect {|s|
                 s.split(/\s*=\s*/)
             }.each {|node, value|
                 raise ProgrammingError, "Blank node expected on the left side of UPDATE assignment instead of '#{bn}'" unless node =~ BN
@@ -325,7 +324,7 @@ class SquishQuery
     # starting keyword, SELECT, INSERT or UPDATE
     attr_reader :key
 
-    # type of query, :select or :merge
+    # type of query, :select or :assert
     attr_reader :type
 
     # blank variables control section
@@ -337,19 +336,28 @@ class SquishQuery
     # literal SQL expression
     attr_reader :literal
 
+    # SQL GROUP BY expression
+    attr_reader :group
+
     # SQL order expression
     attr_reader :order
 
     # direction of order, ASC or DESC
     attr_reader :order_dir
 
+    # number of tuples to return
+    attr_reader :limit
+
+    # number of tuples to skip
+    attr_reader :offset
+
     # query namespaces mapping
     attr_reader :ns
 
-    # replace ${n} substitutions with query string literals
+    # replace 'n' substitutions with query string literals
     #
     def unstring(s)
-        String === s ? s.gsub(/\$\{(\d+)\}/){ @strings[$1.to_i] or $&} : s
+        s.kind_of?(String) ? s.gsub(/'(\d+)'/) { @strings[$1.to_i] or $& } : s
     end
 
     # replace schema uri with namespace prefix
@@ -526,7 +534,7 @@ class SquishQuery
 
     attr_reader :c, :pm, :jc, :b, :table
 
-    # validate expression 
+    # validate expression
     #
     # expression := value [ operator expression ]
     #
@@ -535,16 +543,39 @@ class SquishQuery
     # whitespace between tokens (except inside parentheses) is mandatory
     #
     def validate_expression(string)
-        # todo: complete de-tainting
-        string.split(/[\s()]+/).each do |token|
+        # todo: lexical analyser
+        string.split(/[\s()]+/).collect do |token|
             case token
-            when BN   # blank_node
-            when LITERAL   # literal_string
-            when /\A-?[[:digit:]]+(\.[[:digit:]]+)?\z/   # number
-            when /\A(\+|-|\*|\/|<|<=|>|>=|=|I?LIKE|AND|OR)\z/i   # operator
+            when BN, LITERAL, NUMBER, OPERATOR, AGGREGATE
             else
                 raise ProgrammingError, "Bad token '#{token}' in expression"
             end
         end
     end
+
+    # translate Squish SELECT query to SQL
+    #
+    def to_sql
+        raise ProgrammingError, "Wrong query type: select expected" unless
+            type == :select
+
+        # literal condition is and-ed to the join conditions
+        where = jc.dup
+        where.push '(' + literal + ')' unless literal == ''
+
+        # now put it all together
+        result = 'SELECT ' + nodes.join(', ') +
+            "\nFROM " + table.collect {|a,t| t + ' ' + a }.join(', ') +
+            ((where.length > 0)? "\nWHERE " + where.join("\n  AND ") : '') +
+            ((group == '') ? '' : "\nGROUP BY " + group) +
+            ((order == '') ? '' : "\nORDER BY " + order + ' ' + order_dir) +
+            ((limit == '') ? '' : "\nLIMIT " + limit) +
+            ((offset == '') ? '' : "\nOFFSET " + offset)
+
+        # replace blank node names with bindings
+        result.gsub!(/\?\b.+?\b/) {|node| b[node] or node }
+        raise ProgrammingError, "Unexpected '?' in translated query (probably, caused by unmapped blank node): #{result.gsub(/\s+/, ' ')};" if result =~ /\?/
+
+        unstring(result)
+    end
 end
